# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Time-dependent models."""

import logging
import numpy as np
import scipy.interpolate
from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from gammapy.maps import MapAxis, RegionNDMap, TimeMapAxis
from astropy.utils import lazyproperty
from gammapy.modeling import Parameter
from gammapy.utils.compat import COPY_IF_NEEDED
from gammapy.utils.random import InverseCDFSampler, get_random_state
from gammapy.utils.scripts import make_path
from gammapy.utils.time import time_ref_from_dict, time_ref_to_dict
from .core import ModelBase, _build_parameters_from_dict

__all__ = [
    "ConstantTemporalModel",
    "ExpDecayTemporalModel",
    "GaussianTemporalModel",
    "GeneralizedGaussianTemporalModel",
    "LightCurveTemplateTemporalModel",
    "LinearTemporalModel",
    "PowerLawTemporalModel",
    "SineTemporalModel",
    "TemplatePhaseCurveTemporalModel",
    "TemporalModel",
]

log = logging.getLogger(__name__)


# TODO: make this a small ABC to define a uniform interface.
class TemporalModel(ModelBase):
    """Temporal model base class.

    Evaluates on `~astropy.time.Time` objects.
    """

    _type = "temporal"

    def __init__(self, **kwargs):
        scale = kwargs.pop("scale", "utc")

        if scale not in Time.SCALES:
            raise ValueError(
                f"{scale} is not a valid time scale. Choose from {Time.SCALES}"
            )

        self.scale = scale
        super().__init__(**kwargs)

    def __call__(self, time, energy=None):
        """Evaluate model.

        Parameters
        ----------
        time : `~astropy.time.Time`
            Time object.
        energy : `~astropy.units.Quantity`, optional
            Energy. Default is None.

        Returns
        -------
        values : `~astropy.units.Quantity`
            Model values.
        """
        kwargs = {par.name: par.quantity for par in self.parameters}

        if energy is not None:
            kwargs["energy"] = energy

        time = Time(time, scale=self.scale).mjd * u.d
        return self.evaluate(time, **kwargs)

    @property
    def type(self):
        return self._type

    @property
    def is_energy_dependent(self):
        return False

    @property
    def reference_time(self):
        """Reference time in MJD."""
        return Time(self.t_ref.value, format="mjd", scale=self.scale)

    @reference_time.setter
    def reference_time(self, t_ref):
        """Reference time."""
        if not isinstance(t_ref, Time):
            raise TypeError(f"{t_ref} is not a {Time} object")
        self.t_ref.value = Time(t_ref, scale=self.scale).mjd

    def to_dict(self, full_output=False):
        """Create dictionary for YAML serilisation."""
        data = super().to_dict(full_output)
        data["temporal"]["scale"] = self.scale
        return data

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create a temporal model from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary containing the model parameters.
        **kwargs : dict
            Keyword arguments passed to `~TemporalModel.from_parameters`.
        """
        kwargs = kwargs or {}
        temporal_data = data.get("temporal", data)
        if "scale" in temporal_data:
            kwargs["scale"] = temporal_data["scale"]
        return super().from_dict(data, **kwargs)

    @staticmethod
    def time_sum(t_min, t_max):
        """Total time between t_min and t_max.

        Parameters
        ----------
        t_min, t_max : `~astropy.time.Time`
            Lower and upper bound of integration range.

        Returns
        -------
        time_sum : `~astropy.time.TimeDelta`
            Summed time in the intervals.

        """
        diff = t_max - t_min
        return np.sum(diff).to(u.day)

    def plot(self, time_range, ax=None, n_points=100, **kwargs):
        """
        Plot the temporal model.

        Parameters
        ----------
        time_range : `~astropy.time.Time`
            Times to plot the model.
        ax : `~matplotlib.axes.Axes`, optional
            Axis to plot on.
        n_points : int
            Number of bins to plot model. Default is 100.
        **kwargs : dict
            Keywords forwarded to `~matplotlib.pyplot.errorbar`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes.
        """
        time_min, time_max = time_range
        time_axis = TimeMapAxis.from_time_bounds(
            time_min=time_min, time_max=time_max, nbin=n_points
        )

        m = RegionNDMap.create(region=None, axes=[time_axis])
        kwargs.setdefault("marker", "None")
        kwargs.setdefault("ls", "-")
        kwargs.setdefault("xerr", None)
        m.quantity = self(time_axis.time_mid).to(u.one)
        ax = m.plot(ax=ax, **kwargs)
        ax.set_ylabel("Norm / A.U.")
        return ax

    def sample_time(self, n_events, t_min, t_max, t_delta="1 s", random_state=0):
        """Sample arrival times of events.

        Parameters
        ----------
        n_events : int
            Number of events to sample.
        t_min : `~astropy.time.Time`
            Start time of the sampling.
        t_max : `~astropy.time.Time`
            Stop time of the sampling.
        t_delta : `~astropy.units.Quantity`, optional
            Time step used for sampling of the temporal model. Default is 1 s.
        random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}
            Defines random number generator initialisation.
            Passed to `~gammapy.utils.random.get_random_state`.
            Default is 0.

        Returns
        -------
        time : `~astropy.units.Quantity`
            Array with times of the sampled events.
        """
        t_min = Time(t_min, scale=self.scale)
        t_max = Time(t_max, scale=self.scale)
        t_delta = u.Quantity(t_delta)
        random_state = get_random_state(random_state)

        ontime = (t_max - t_min).to("s")
        n_step = np.ceil((ontime / t_delta).to_value("")).item()
        t_step = ontime / n_step

        # take n+1 bins
        indices = np.arange(n_step + 1)

        # evaluate model at bin center
        centers = (indices[:-1] + 0.5) * t_step
        t = Time(t_min + centers, format="mjd")
        pdf = self(t)

        # build sample list
        sampler = InverseCDFSampler(pdf=pdf, random_state=random_state)
        time_pix = sampler.sample(n_events)[0]

        # transform bin in time after shift by half a pixel
        steps = indices * t_step
        time = np.interp(time_pix + 0.5, indices, steps)
        return t_min + time

    def integral(self, t_min, t_max, oversampling_factor=100, **kwargs):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min: `~astropy.time.Time`
            Start times of observation.
        t_max: `~astropy.time.Time`
            Stop times of observation.
        oversampling_factor : int, optional
            Oversampling factor to be used for numerical integration.
            Default is 100.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        t_values, steps = np.linspace(
            t_min.mjd, t_max.mjd, oversampling_factor, retstep=True, axis=-1
        )
        times = Time(t_values, format="mjd", scale=self.scale)
        values = self(times)
        integral = np.sum(values, axis=-1) * steps
        return integral / self.time_sum(t_min, t_max).to_value("d")


class ConstantTemporalModel(TemporalModel):
    """Constant temporal model.

    For more information see :ref:`constant-temporal-model`.
    """

    tag = ["ConstantTemporalModel", "const"]

    @staticmethod
    def evaluate(time):
        """Evaluate at given times."""
        return np.ones(time.shape) * u.one

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min : `~astropy.time.Time`
            Start times of observation.
        t_max : `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : `~astropy.units.Quantity`
            Integrated flux norm on the given time intervals.
        """
        return (t_max - t_min) / self.time_sum(t_min, t_max)


class LinearTemporalModel(TemporalModel):
    """Temporal model with a linear variation.

    For more information see :ref:`linear-temporal-model`.

    Parameters
    ----------
    alpha : float
        Constant term of the baseline flux.
        Default is 1.
    beta : `~astropy.units.Quantity`
        Time variation coefficient of the flux.
        Default is 0.
    t_ref : `~astropy.units.Quantity`
        The reference time in mjd.
        Frozen per default, at 2000-01-01.
    """

    tag = ["LinearTemporalModel", "linear"]

    alpha = Parameter("alpha", 1.0, frozen=False)
    beta = Parameter("beta", 0.0, unit="d-1", frozen=False)
    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=True)

    @staticmethod
    def evaluate(time, alpha, beta, t_ref):
        """Evaluate at given times."""
        return alpha + beta * (time - t_ref)

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min : `~astropy.time.Time`
            Start times of observation.
        t_max : `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        pars = self.parameters
        alpha = pars["alpha"]
        beta = pars["beta"].quantity
        t_ref = self.reference_time
        value = alpha * (t_max - t_min) + beta / 2.0 * (
            (t_max - t_ref) * (t_max - t_ref) - (t_min - t_ref) * (t_min - t_ref)
        )
        return value / self.time_sum(t_min, t_max)


class ExpDecayTemporalModel(TemporalModel):
    r"""Temporal model with an exponential decay.

    For more information see :ref:`expdecay-temporal-model`.

    Parameters
    ----------
    t0 : `~astropy.units.Quantity`
        Decay timescale. Default is 1 day.
    t_ref : `~astropy.units.Quantity`
        The reference time in mjd. Frozen per default, at 2000-01-01.
    """

    tag = ["ExpDecayTemporalModel", "exp-decay"]

    t0 = Parameter("t0", "1 d", frozen=False)
    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=True)

    @staticmethod
    def evaluate(time, t0, t_ref):
        """Evaluate at given times."""
        return np.exp(-(time - t_ref) / t0)

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min : `~astropy.time.Time`
            Start times of observation.
        t_max : `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        pars = self.parameters
        t0 = pars["t0"].quantity
        t_ref = self.reference_time
        value = self.evaluate(t_max, t0, t_ref) - self.evaluate(t_min, t0, t_ref)
        return -t0 * value / self.time_sum(t_min, t_max)


class GaussianTemporalModel(TemporalModel):
    r"""A Gaussian temporal profile.

    For more information see :ref:`gaussian-temporal-model`.

    Parameters
    ----------
    t_ref : `~astropy.units.Quantity`
        The reference time in mjd at the peak.
        Default is 2000-01-01.
    sigma : `~astropy.units.Quantity`
        Width of the gaussian profile.
        Default is 1 day.
    """

    tag = ["GaussianTemporalModel", "gauss"]

    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=False)
    sigma = Parameter("sigma", "1 d", frozen=False)

    @staticmethod
    def evaluate(time, t_ref, sigma):
        return np.exp(-((time - t_ref) ** 2) / (2 * sigma**2))

    def integral(self, t_min, t_max, **kwargs):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min : `~astropy.time.Time`
            Start times of observation.
        t_max : `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        pars = self.parameters
        sigma = pars["sigma"].quantity
        t_ref = self.reference_time
        norm = np.sqrt(np.pi / 2) * sigma

        u_min = (t_min - t_ref) / (np.sqrt(2) * sigma)
        u_max = (t_max - t_ref) / (np.sqrt(2) * sigma)

        integral = norm * (scipy.special.erf(u_max) - scipy.special.erf(u_min))
        return integral / self.time_sum(t_min, t_max)


class GeneralizedGaussianTemporalModel(TemporalModel):
    r"""A generalized Gaussian temporal profile.

    For more information see :ref:`generalized-gaussian-temporal-model`.

    Parameters
    ----------
    t_ref : `~astropy.units.Quantity`
        The time of the pulse's maximum intensity.
        Default is 2000-01-01.
    t_rise : `~astropy.units.Quantity`
        Rise time constant.
        Default is 1 day.
    t_decay : `~astropy.units.Quantity`
        Decay time constant.
        Default is 1 day.
    eta : `~astropy.units.Quantity`
        Inverse pulse sharpness -> higher values implies a more peaked pulse.
        Default is 1/2.

    """

    tag = ["GeneralizedGaussianTemporalModel", "gengauss"]

    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=False)
    t_rise = Parameter("t_rise", "1d", frozen=False)
    t_decay = Parameter("t_decay", "1d", frozen=False)
    eta = Parameter("eta", 1 / 2, unit="", frozen=False)

    @staticmethod
    def evaluate(time, t_ref, t_rise, t_decay, eta):
        val_rise = np.exp(
            -0.5
            * (np.abs(u.Quantity(time - t_ref, "d")) ** (1 / eta))
            / (t_rise ** (1 / eta))
        )
        val_decay = np.exp(
            -0.5
            * (np.abs(u.Quantity(time - t_ref, "d")) ** (1 / eta))
            / (t_decay ** (1 / eta))
        )
        val = np.where(time < t_ref, val_rise, val_decay)
        return val


class LightCurveTemplateTemporalModel(TemporalModel):
    """Temporal light curve model.

    The lightcurve is given at specific times (and optionally energies) as a ``norm``
    It can be serialised either as an astropy table or a `~gammapy.maps.RegionNDMap`

    The ``norm`` is supposed to be a unit-less multiplicative factor in the model,
    to be multiplied with a spectral model.

    The model does linear interpolation for times between the given ``(time, energy, norm)``
    values.

    When the temporal model is energy-dependent, the default interpolation scheme is
    linear with a log scale for the values. The interpolation method and scale values
    can be changed with the ``method`` and ``values_scale`` arguments.

    For more information see :ref:`LightCurve-temporal-model`.

    Examples
    --------
    Read an example light curve object:

    >>> from gammapy.modeling.models import LightCurveTemplateTemporalModel
    >>> path = '$GAMMAPY_DATA/tests/models/light_curve/lightcrv_PKSB1222+216.fits'
    >>> light_curve = LightCurveTemplateTemporalModel.read(path)

    Show basic information about the lightcurve:

    >>> print(light_curve)
    LightCurveTemplateTemporalModel model summary:
     Reference time: 59000.49919925926 MJD
     Start time: 58999.99919925926 MJD
     End time: 61862.99919925926 MJD
     Norm min: 0.01551196351647377
    Norm max: 1.0

    <BLANKLINE>

    Compute ``norm`` at a given time:

    >>> from astropy.time import Time
    >>> t = Time(59001.195, format="mjd")
    >>> light_curve.evaluate(t)
    <Quantity [0.02288737]>

    Compute mean ``norm`` in a given time interval:

    >>> import astropy.units as u
    >>> t_r = Time(59000.5, format='mjd')
    >>> t_min = t_r + [1, 4, 8] * u.d
    >>> t_max = t_r + [1.5, 6, 9] * u.d
    >>> light_curve.integral(t_min, t_max)
    <Quantity [0.00375698, 0.0143724 , 0.00688029]>
    """

    tag = ["LightCurveTemplateTemporalModel", "template"]

    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=True)

    def __init__(self, map, t_ref=None, filename=None, method=None, values_scale=None):
        if (map.data < 0).any():
            log.warning("Map has negative values. Check and fix this!")

        self.map = map.copy()
        super().__init__()

        if t_ref:
            self.reference_time = t_ref

        self.filename = filename

        if method is None:
            method = "linear"

        if values_scale is None:
            if self.is_energy_dependent:
                values_scale = "log"
            else:
                values_scale = "lin"

        self.method = method
        self.values_scale = values_scale

    def __str__(self):
        start_time = self.t_ref.quantity + self.map.geom.axes["time"].edges[0]
        end_time = self.t_ref.quantity + self.map.geom.axes["time"].edges[-1]
        norm_min = np.nanmin(self.map.data)
        norm_max = np.nanmax(self.map.data)

        prnt = (
            f"{self.__class__.__name__} model summary:\n "
            f"Reference time: {self.t_ref.value} MJD \n "
            f"Start time: {start_time.value} MJD \n "
            f"End time: {end_time.value} MJD \n "
            f"Norm min: {norm_min} \n"
            f"Norm max: {norm_max}"
        )

        if self.is_energy_dependent:
            energy_min = self.map.geom.axes["energy"].center[0]
            energy_max = self.map.geom.axes["energy"].center[-1]
            prnt1 = f"Energy min: {energy_min} \n" f"Energy max: {energy_max} \n"
            prnt = prnt + prnt1

        return prnt

    @property
    def is_energy_dependent(self):
        """Whether the model is energy dependent."""
        return self.map.geom.has_energy_axis

    @classmethod
    def from_table(cls, table, filename=None):
        """Create a template model from an astropy table.

        Parameters
        ----------
        table : `~astropy.table.Table`
            Table containing the template model.
        filename : str, optional
            Name of input file. Default is None.

        Returns
        -------
        model : `LightCurveTemplateTemporalModel`
            Light curve template model.
        """
        columns = [_.lower() for _ in table.colnames]
        if "time" not in columns:
            raise ValueError("A TIME column is necessary")

        t_ref = time_ref_from_dict(table.meta, scale="utc")
        nodes = table["TIME"]

        ax_unit = nodes.quantity.unit

        if not ax_unit.is_equivalent("d"):
            try:
                ax_unit = u.Unit(table.meta["TIMEUNIT"])
            except KeyError:
                raise ValueError("Time unit not found in the table")

        time_axis = MapAxis.from_nodes(nodes=nodes, name="time", unit=ax_unit)
        axes = [time_axis]
        m = RegionNDMap.create(region=None, axes=axes, data=table["NORM"])

        return cls(m, t_ref=t_ref, filename=filename)

    @classmethod
    def read(cls, filename, format="table"):
        """Read a template model.

        Parameters
        ----------
        filename : str
            Name of file to read.
        format : {"table", "map"}
            Format of the input file.

        Returns
        -------
        model : `LightCurveTemplateTemporalModel`
            Light curve template model.
        """
        filename = str(make_path(filename))
        if format == "table":
            table = Table.read(filename)
            return cls.from_table(table, filename=filename)

        elif format == "map":
            with fits.open(filename) as hdulist:
                header = hdulist["SKYMAP_BANDS"].header
                t_ref = time_ref_from_dict(header)
                # TODO : Ugly hack to prevent creating a TimeMapAxis
                # By default, MapAxis.from_table tries to create a
                # TimeMapAxis, failing which, it creates a normal MapAxis.
                # This ugly hack forces the fail. We need a normal Axis to
                # have the evaluate method work
                hdulist["SKYMAP_BANDS"].header.pop("MJDREFI")
                m = RegionNDMap.from_hdulist(hdulist)
            return cls(m, t_ref=t_ref, filename=filename)

        else:
            raise ValueError(
                f"Not a valid format: '{format}', choose from: {'table', 'map'}"
            )

    def to_table(self):
        """Convert model to an astropy table."""
        if self.is_energy_dependent:
            raise NotImplementedError("Not supported for energy dependent models")
        table = Table(
            data=[self.map.geom.axes["time"].center, self.map.quantity],
            names=["TIME", "NORM"],
            meta=time_ref_to_dict(self.reference_time, scale=self.scale),
        )
        return table

    def write(self, filename=None, format="table", overwrite=False):
        """Write a model to disk as per the specified format.

        Parameters
        ----------
        filename : str, optional
            Name of output file. By default, the template model will be saved
            with the `LightCurveTemplateTemporalModel.filename` attribute.
            If `filename` is provided, this attribute will be used.
        format : {"table" or "map"}
            If format is "table", it is serialised as a `~astropy.table.Table`.
            If "map", then it is serialised as a `~gammapy.maps.RegionNDMap`.
            Default is "table".
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        """
        if filename is not None:
            self.filename = filename

        if self.filename is None:
            raise IOError("Missing filename")

        if format == "table":
            table = self.to_table()
            table.write(self.filename, overwrite=overwrite)
        elif format == "map":
            # RegionNDMap.from_hdulist does not update the header
            hdulist = self.map.to_hdulist()
            hdulist["SKYMAP_BANDS"].header.update(
                time_ref_to_dict(self.reference_time, scale=self.scale)
            )
            hdulist.writeto(self.filename, overwrite=overwrite)
        else:
            raise ValueError("Not a valid format, choose from ['map', 'table']")

    def evaluate(self, time, t_ref=None, energy=None):
        """Evaluate the model at given coordinates.

        Parameters
        ----------
        time: `~astropy.time.Time`
            Time.
        t_ref: `~gammapy.modeling.Parameter`, optional
            Reference time for the model. Default is None.
        energy: `~astropy.units.Quantity`, optional
            Energy. Default is None.

        Returns
        -------
        values : `~astropy.units.Quantity`
            Model values.
        """
        if t_ref is None:
            t_ref = self.reference_time

        t = (time - t_ref).to_value(self.map.geom.axes["time"].unit)
        coords = {"time": t}

        if self.is_energy_dependent:
            if energy is None:
                energy = self.map.geom.axes["energy"].center

            coords["energy"] = energy.reshape(-1, 1)

        val = self.map.interp_by_coord(
            coords, method=self.method, values_scale=self.values_scale
        )
        val = np.clip(val, 0, a_max=None)
        return u.Quantity(val, unit=self.map.unit, copy=COPY_IF_NEEDED)

    def integral(self, t_min, t_max, oversampling_factor=100, **kwargs):
        if self.is_energy_dependent:
            raise NotImplementedError(
                "Integral not supported for energy dependent models"
            )

        return super().integral(t_min, t_max, oversampling_factor, **kwargs)

    @classmethod
    def from_dict(cls, data):
        data = data["temporal"]
        filename = data["filename"]
        format = data.get("format", "table")
        return cls.read(filename, format)

    def _guess_format(self):
        if self.is_energy_dependent:
            format = "map"
        else:
            format = "table"
        log.info("Inferred format: " + format)
        return format

    def to_dict(self, full_output=False, format=None):
        """Create dictionary for YAML serialisation."""
        data = super().to_dict(full_output)
        if format is None:
            format = self._guess_format()
        data["temporal"]["filename"] = self.filename
        data["temporal"]["format"] = format
        data["temporal"]["unit"] = str(self.map.unit)
        return data

    def plot(self, time_range, ax=None, n_points=100, energy=None, **kwargs):
        """
        Plot the temporal model.

        Parameters
        ----------
        time_range : `~astropy.time.Time`
            Times to plot the model.
        ax : `~matplotlib.axes.Axes`, optional
            Axis to plot on. Default is None.
        n_points : int, optional
            Number of bins to plot model. Default is 100.
        energy : `~astropy.units.quantity`, optional
            Energies to compute the model at for energy dependent models. Default is None.
        **kwargs : dict
            Keywords forwarded to `~matplotlib.pyplot.errorbar`.
        Returns
        -------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes.
        """
        if not self.is_energy_dependent:
            super().plot(time_range=time_range, ax=ax, n_points=n_points, **kwargs)
        else:
            time_min, time_max = Time(time_range, scale=self.scale)
            time_axis = TimeMapAxis.from_time_bounds(
                time_min=time_min, time_max=time_max, nbin=n_points
            )
            if energy is None:
                energy_axis = self.map.geom.axes["energy"]
            else:
                energy_axis = MapAxis.from_nodes(
                    nodes=energy, name="energy", interp="log"
                )

            m = RegionNDMap.create(region=None, axes=[time_axis, energy_axis])
            kwargs.setdefault("marker", "None")
            kwargs.setdefault("ls", "-")
            m.quantity = self.evaluate(
                time=time_axis.time_mid, energy=energy_axis.center
            )
            ax = m.plot(axis_name="time", ax=ax, **kwargs)
            ax.set_ylabel("Norm / A.U.")

            return ax, m


class PowerLawTemporalModel(TemporalModel):
    """Temporal model with a Power Law decay.

    For more information see :ref:`powerlaw-temporal-model`.

    Parameters
    ----------
    alpha : float
        Decay time power. Default is 1.
    t_ref: `~astropy.units.Quantity`
        The reference time in mjd.
        Frozen by default, at 2000-01-01.
    t0: `~astropy.units.Quantity`
        The scaling time in mjd.
        Fixed by default, at 1 day.
    """

    tag = ["PowerLawTemporalModel", "powerlaw"]

    alpha = Parameter("alpha", 1.0, frozen=False)
    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=True)
    t0 = Parameter("t0", "1 d", frozen=True)

    @staticmethod
    def evaluate(time, alpha, t_ref, t0=1 * u.day):
        """Evaluate at given times."""
        return np.power((time - t_ref) / t0, alpha)

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min: `~astropy.time.Time`
            Start times of observation.
        t_max: `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        pars = self.parameters
        alpha = pars["alpha"].quantity
        t0 = pars["t0"].quantity
        t_ref = self.reference_time
        if alpha != -1:
            value = self.evaluate(t_max, alpha + 1.0, t_ref, t0) - self.evaluate(
                t_min, alpha + 1.0, t_ref, t0
            )
            return t0 / (alpha + 1.0) * value / self.time_sum(t_min, t_max)
        else:
            value = np.log((t_max - t_ref) / (t_min - t_ref))
            return t0 * value / self.time_sum(t_min, t_max)


class SineTemporalModel(TemporalModel):
    """Temporal model with a sinusoidal modulation.

    For more information see :ref:`sine-temporal-model`.

    Parameters
    ----------
    amp : float
        Amplitude of the sinusoidal function.
        Default is 1.
    t_ref: `~astropy.units.Quantity`
        The reference time in mjd.
        Default is 2000-01-01.
    omega: `~astropy.units.Quantity`
        Pulsation of the signal.
        Default is 1 rad/day.
    """

    tag = ["SineTemporalModel", "sinus"]

    amp = Parameter("amp", 1.0, frozen=False)
    omega = Parameter("omega", "1. rad/day", frozen=False)
    _t_ref_default = Time("2000-01-01")
    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=False)

    @staticmethod
    def evaluate(time, amp, omega, t_ref):
        """Evaluate at given times."""
        return 1.0 + amp * np.sin(omega * (time - t_ref))

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min: `~astropy.time.Time`
            Start times of observation.
        t_max: `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm : float
            Integrated flux norm on the given time intervals.
        """
        pars = self.parameters
        omega = pars["omega"].quantity.to_value("rad/day")
        amp = pars["amp"].value
        t_ref = self.reference_time

        value = (t_max - t_min).to_value(u.day) - amp / omega * (
            np.sin(omega * (t_max - t_ref).to_value(u.day))
            - np.sin(omega * (t_min - t_ref).to_value(u.day))
        )
        return value / self.time_sum(t_min, t_max).to_value(u.day)


class TemplatePhaseCurveTemporalModel(TemporalModel):
    """Temporal phase curve model.

    A timing solution is used to compute the phase corresponding to time and
    a template phase curve is used to determine the associated ``norm``.

    The phasecurve is given as a table with columns ``phase`` and ``norm``.

    The ``norm`` is supposed to be a unit-less multiplicative factor in the model,
    to be multiplied with a spectral model.

    The model does linear interpolation for times between the given ``(phase, norm)`` values.

    The implementation currently uses `scipy.interpolate. InterpolatedUnivariateSpline`,
    using degree ``k=1`` to get linear interpolation.
    This class also contains an ``integral`` method, making the computation of
    mean fluxes for a given time interval a one-liner.

    Parameters
    ----------
    table : `~astropy.table.Table`
        A table with 'PHASE' vs 'NORM'.
    filename : str
        The name of the file containing the phase curve.
    t_ref : `~astropy.units.Quantity`
        The reference time in mjd.
        Default is 48442.5 mjd.
    phi_ref : `~astropy.units.Quantity`
        The phase at reference time.
        Default is 0.
    f0 : `~astropy.units.Quantity`
        The frequency at t_ref in s-1.
        Default is 29.946923 s-1.
    f1 : `~astropy.units.Quantity`
        The frequency derivative at t_ref in s-2.
        Default is 0 s-2.
    f2 : `~astropy.units.Quantity`
        The frequency second derivative at t_ref in s-3.
        Default is 0 s-3.
    """

    tag = ["TemplatePhaseCurveTemporalModel", "template-phase"]
    _t_ref_default = Time(48442.5, format="mjd")
    _phi_ref_default = 0
    _f0_default = 29.946923 * u.s**-1
    _f1_default = 0 * u.s**-2
    _f2_default = 0 * u.s**-3

    t_ref = Parameter("t_ref", _t_ref_default.mjd, unit="day", frozen=True)
    phi_ref = Parameter("phi_ref", _phi_ref_default, unit="", frozen=True)
    f0 = Parameter("f0", _f0_default, frozen=True)
    f1 = Parameter("f1", _f1_default, frozen=True)
    f2 = Parameter("f2", _f2_default, frozen=True)

    def __init__(self, table, filename=None, **kwargs):
        self.table = self._normalise_table(table)
        if filename is not None:
            filename = str(make_path(filename))
        if filename is None:
            log.warning(
                "The filename is not defined. Therefore, the model will not be serialised correctly. "
                'To set the filename, the "template_model.filename" attribute can be used.'
            )
        self.filename = filename
        super().__init__(**kwargs)

    @staticmethod
    def _normalise_table(table):
        x = table["PHASE"].data
        y = table["NORM"].data

        interpolator = scipy.interpolate.InterpolatedUnivariateSpline(
            x, y, k=1, ext=2, bbox=[0.0, 1.0]
        )

        integral = interpolator.integral(0, 1)

        table["NORM"] *= 1 / integral
        return table

    @classmethod
    def read(
        cls,
        path,
        t_ref=_t_ref_default.mjd * u.d,
        phi_ref=_phi_ref_default,
        f0=_f0_default,
        f1=_f1_default,
        f2=_f2_default,
    ):
        """Read phasecurve model table from FITS file.

        Beware : this does **not** read parameters.
        They will be set to defaults.

        Parameters
        ----------
        path : str or `~pathlib.Path`
            Filename with path.
        """
        filename = str(make_path(path))

        return cls(
            Table.read(filename),
            filename=filename,
            t_ref=t_ref,
            phi_ref=phi_ref,
            f0=f0,
            f1=f1,
            f2=f2,
        )

    @staticmethod
    def _time_to_phase(time, t_ref, phi_ref, f0, f1, f2):
        """Convert time to phase given timing solution parameters.

        Parameters
        ----------
        time : `~astropy.units.Quantity`
            The time at which to compute the phase.
        t_ref : `~astropy.units.Quantity`
            The reference time in mjd.
        phi_ref : `~astropy.units.Quantity`
            The phase at reference time.
            Default is 0.
        f0 : `~astropy.units.Quantity`
            The frequency at t_ref in s-1.
        f1 : `~astropy.units.Quantity`
            The frequency derivative at t_ref in s-2.
        f2 : `~astropy.units.Quantity`
            The frequency second derivative at t_ref in s-3.

        Returns
        -------
        phase : float
            Phase.
        period_number : int
            Number of period since t_ref.
        """
        delta_t = time - t_ref
        phase = (
            phi_ref + delta_t * (f0 + delta_t / 2.0 * (f1 + delta_t / 3 * f2))
        ).to_value("")

        period_number = np.floor(phase)
        phase -= period_number
        return phase, period_number

    def write(self, path=None, overwrite=False):
        if path is None:
            path = self.filename
        if path is None:
            raise ValueError(f"filename is required for {self.tag}")
        else:
            self.filename = str(make_path(path))
            self.table.write(self.filename, overwrite=overwrite)

    @lazyproperty
    def _interpolator(self):
        x = self.table["PHASE"].data
        y = self.table["NORM"].data

        return scipy.interpolate.InterpolatedUnivariateSpline(
            x, y, k=1, ext=2, bbox=[0.0, 1.0]
        )

    def evaluate(self, time, t_ref, phi_ref, f0, f1, f2):
        phase, _ = self._time_to_phase(time, t_ref, phi_ref, f0, f1, f2)
        return self._interpolator(phase) * u.one

    def integral(self, t_min, t_max):
        """Evaluate the integrated flux within the given time intervals.

        Parameters
        ----------
        t_min: `~astropy.time.Time`
            Start times of observation.
        t_max: `~astropy.time.Time`
            Stop times of observation.

        Returns
        -------
        norm: The model integrated flux.
        """
        kwargs = {par.name: par.quantity for par in self.parameters}
        ph_min, n_min = self._time_to_phase(t_min.mjd * u.d, **kwargs)
        ph_max, n_max = self._time_to_phase(t_max.mjd * u.d, **kwargs)

        # here we assume that the frequency does not change during the integration boundaries
        delta_t = (t_min - self.reference_time).to(u.d)
        frequency = self.f0.quantity + delta_t * (
            self.f1.quantity + delta_t * self.f2.quantity / 2
        )

        # Compute integral of one phase
        phase_integral = self._interpolator.antiderivative()(
            1
        ) - self._interpolator.antiderivative()(0)
        # Multiply by the total number of phases
        phase_integral *= n_max - n_min - 1

        # Compute integrals before first full phase and after the last full phase
        end_integral = self._interpolator.antiderivative()(
            ph_max
        ) - self._interpolator.antiderivative()(0)
        start_integral = self._interpolator.antiderivative()(
            1
        ) - self._interpolator.antiderivative()(ph_min)

        # Divide by Jacobian (here we neglect variations of frequency during the integration period)
        total = phase_integral + start_integral + end_integral
        # Normalize by total integration time
        n_period = (self.time_sum(t_min, t_max) * frequency).to("")
        if int(n_period) == 0:
            n_period = 1

        integral_norm = total / n_period

        return integral_norm

    @classmethod
    def from_dict(cls, data):
        params = _build_parameters_from_dict(
            data["temporal"]["parameters"], cls.default_parameters
        )
        filename = data["temporal"]["filename"]
        kwargs = {par.name: par for par in params}
        return cls.read(filename, **kwargs)

    def to_dict(self, full_output=False):
        """Create dictionary for YAML serialisation."""
        model_dict = super().to_dict()
        model_dict["temporal"]["filename"] = self.filename
        return model_dict

    def plot_phasogram(self, ax=None, n_points=100, **kwargs):
        """
        Plot phasogram of the phase model.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axis to plot on. Default is None.
        n_points : int, optional
            Number of bins to plot model. Default is 100.
        **kwargs : dict
            Keywords forwarded to `~matplotlib.pyplot.errorbar`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes.
        """
        phase_axis = MapAxis.from_bounds(0.0, 1, nbin=n_points, name="Phase", unit="")

        m = RegionNDMap.create(region=None, axes=[phase_axis])
        kwargs.setdefault("marker", "None")
        kwargs.setdefault("ls", "-")
        kwargs.setdefault("xerr", None)
        m.quantity = self._interpolator(phase_axis.center)
        ax = m.plot(ax=ax, **kwargs)
        ax.set_ylabel("Norm / A.U.")
        return ax

    def sample_time(self, n_events, t_min, t_max, t_delta="1 s", random_state=0):
        """Sample arrival times of events.

        To fully cover the phase range, t_delta is the minimum between the input
        and product of the period at 0.5*(t_min + t_max) and the table bin size.

        Parameters
        ----------
        n_events : int
            Number of events to sample.
        t_min : `~astropy.time.Time`
            Start time of the sampling.
        t_max : `~astropy.time.Time`
            Stop time of the sampling.
        t_delta : `~astropy.units.Quantity`
            Time step used for sampling of the temporal model.
        random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}
            Defines random number generator initialisation.
            Passed to `~gammapy.utils.random.get_random_state`.

        Returns
        -------
        time : `~astropy.units.Quantity`
            Array with times of the sampled events.
        """
        t_delta = u.Quantity(t_delta)

        # Determine period at the mid time
        t_mid = Time(t_min, scale=self.scale) + 0.5 * (t_max - t_min)
        delta_t = (t_mid - self.reference_time).to(u.d)
        frequency = self.f0.quantity + delta_t * (
            self.f1.quantity + delta_t * self.f2.quantity / 2
        )
        period = 1 / frequency

        # Take minimum time delta between user input and the period divided by the number of rows in the model table
        # this assumes that phase values are evenly spaced.
        t_delta = np.minimum(period / len(self.table), t_delta)

        return super().sample_time(n_events, t_min, t_max, t_delta, random_state)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities for caching"""

import functools
import inspect
import hashlib
import weakref
from gammapy.utils.parallel import is_ray_available

USE_INSTANCE_CACHE = False

if is_ray_available():
    import ray

    def _hash(value):
        try:
            return hash(value)
        except TypeError:
            data = ray.cloudpickle.dumps(value)
            return hashlib.sha256(data).hexdigest()
else:

    def _hash(value):
        return hash(value)


def make_key(sig, *args, **kwargs):
    """
    Generate a unique hash key for caching based on normalized constructor arguments.

    This function uses the signature of the class constructor (`__init__`) to normalize
    the input arguments, serializes them using `ray.cloudpickle`, and returns a SHA-256 hash
    digest. The resulting key is suitable for use in caching mechanisms.

    Parameters
    ----------
    sig : inspect.Signature
        The signature of the method or function used to normalize arguments.
    *args : tuple
        Positional arguments to be normalized.
    **kwargs : dict
        Keyword arguments to be normalized.

    Returns
    -------
    key : str
        A SHA-256 hexadecimal digest representing the normalized arguments.
    """
    bound = sig.bind_partial(None, *args, **kwargs)
    bound.apply_defaults()
    normalized_args = dict(sorted(bound.arguments.items()))
    normalized_args.pop("self", None)
    return _hash(tuple(normalized_args.items()))


class _WeakIdDict:
    """Like `weakref.WeakKeyDictionary`, but uses identity-based hashing and equality.
    from https://github.com/python/cpython/issues/102618#issuecomment-2839489762.
    """

    __slots__ = ("_dict",)

    def __init__(self):
        self._dict = {}

    def __getitem__(self, key):
        item, _ = self._dict[id(key)]
        return item

    def __setitem__(self, key, value):
        id_key = id(key)
        ref = weakref.ref(key, lambda _: self._dict.pop(id_key))
        self._dict[id_key] = (value, ref)


def cachemethod(fn):
    """
    Decorator to cache method results on a per-instance basis using weak references.

    This decorator stores cached results of method calls in a `WeakKeyDictionary`,
    ensuring that the cache is automatically cleared when the instance is garbage collected.
    The cache key is generated using `make_key`, which normalizes and hashes the method
    arguments to ensure consistent and order-independent caching.

    Parameters
    ----------
    fn : callable
        The method to be decorated.

    Returns
    -------
    wrapper : callable
        The wrapped method with caching behavior.
    """

    cache1 = _WeakIdDict()
    sig = inspect.signature(fn)

    parameters = list(sig.parameters.values())
    if len(parameters) == 0 or parameters[0].kind not in {
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
    }:
        raise ValueError(
            "The `@cachemethod` decorator can only be used on instance methods "
            "with a positional `self` argument."
        )

    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        argkey = make_key(sig, *args, **kwargs)
        try:
            out = cache1[self][argkey]
        except KeyError:
            try:
                cache2 = cache1[self]
            except KeyError:
                cache2 = cache1[self] = {}
            out = cache2[argkey] = fn(self, *args, **kwargs)
        return out

    return wrapper


class CacheInstanceMixin:
    """Cache class instance based on init arguments equivalence"""

    def __new__(cls, *args, **kwargs):
        new = super().__new__(cls)
        if not USE_INSTANCE_CACHE or not is_ray_available():
            return new

        # Ensure each subclass has its own cache
        if not hasattr(cls, "_instances"):
            cls._instances = _WeakIdDict()

        try:
            new.__init__(*args, **kwargs)
        except TypeError:
            return new  # ignore cache for unpickle

        sig = inspect.signature(cls.__init__)
        argkey = make_key(sig, *args, **kwargs)
        try:
            out = cls._instances[cls][argkey]
        except KeyError:
            try:
                cache2 = cls._instances[cls]
            except KeyError:
                cache2 = cls._instances[cls] = {}
            out = cache2[argkey] = new
        return out



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pickle
import numpy as np
import gammapy.utils.cache as cache
from gammapy.utils.testing import requires_dependency


class Dummy(cache.CacheInstanceMixin):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __eq__(self, other):
        return isinstance(other, Dummy) and self.a == other.a and self.b == other.b


class Dummy2(cache.CacheInstanceMixin):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __eq__(self, other):
        return isinstance(other, Dummy2) and self.a == other.a and self.b == other.b


@requires_dependency("ray")
def test_dummy_cache():
    cache.USE_INSTANCE_CACHE = True

    x = Dummy(1, np.array(2))
    y = Dummy(1, np.array(2))
    z = Dummy2(1, np.array(2))

    u = Dummy(1, np.array(3))
    v = Dummy2(1, np.array(3))
    w = Dummy2(1, np.array(3))

    assert x is y
    assert z is not y

    assert u is not x
    assert u is not v
    assert v is not z
    assert v is w

    # the cache is not propagated through pickle
    # (implementing that would be more complex)
    data = pickle.dumps(x)
    xnew = pickle.loads(data)
    assert xnew is not x

    cache.USE_INSTANCE_CACHE = False

    x = Dummy(1, np.array(2))
    y = Dummy(1, np.array(2))
    z = Dummy2(1, np.array(2))

    assert x is not y
    assert z is not y



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import collections.abc
import copy
import html
import logging
from os.path import split
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
import matplotlib.pyplot as plt
from gammapy.maps import Map, RegionGeom
from gammapy.modeling import Covariance, Parameter, Parameters
from gammapy.modeling.covariance import CovarianceMixin
from gammapy.utils.scripts import from_yaml, make_path, to_yaml, write_yaml

__all__ = ["Model", "Models", "DatasetModels", "ModelBase"]


log = logging.getLogger(__name__)


def _recursive_dict_filename_update(dict_, path):
    """Update model filename to full path if exits."""
    for key, value in dict_.items():
        if isinstance(value, dict):
            _recursive_dict_filename_update(value, path)
        elif key == "filename":
            filename = dict_[key]
            if (path / filename).exists():
                dict_[key] = path / filename


def _recursive_model_filename_update(model, path):
    """Update model filename to relative path if child of path."""
    if hasattr(model, "filename") and path == make_path(model.filename).parent:
        _, filename = split(model.filename)
        model.filename = filename

    if hasattr(model, "_models"):
        for m in model._models:
            _recursive_model_filename_update(m, path)


def _set_model_link(shared_register, model):
    for param in model.parameters:
        name = param.name
        link_label = param._link_label_io
        if link_label is not None:
            if link_label in shared_register:
                new_param = shared_register[link_label]
                setattr(model, name, new_param)
            else:
                shared_register[link_label] = param
    return shared_register


def _flatten_models(models):
    flat_models = []
    for model in models:
        if hasattr(model, "_models"):
            flat_models.extend(_flatten_models(model._models))
        else:
            flat_models.append(model)
    return flat_models


def _set_models_link(models):
    shared_register = {}
    flat_models = _flatten_models(models)
    for model in flat_models:
        shared_register = _set_model_link(shared_register, model)


def _get_model_class_from_dict(data):
    """Get a model class from a dictionary."""
    from . import (
        MODEL_REGISTRY,
        PRIOR_REGISTRY,
        SPATIAL_MODEL_REGISTRY,
        SPECTRAL_MODEL_REGISTRY,
        TEMPORAL_MODEL_REGISTRY,
    )

    if "type" in data:
        cls = MODEL_REGISTRY.get_cls(data["type"])
    elif "spatial" in data:
        cls = SPATIAL_MODEL_REGISTRY.get_cls(data["spatial"]["type"])
    elif "spectral" in data:
        cls = SPECTRAL_MODEL_REGISTRY.get_cls(data["spectral"]["type"])
    elif "temporal" in data:
        cls = TEMPORAL_MODEL_REGISTRY.get_cls(data["temporal"]["type"])
    elif "prior" in data:
        cls = PRIOR_REGISTRY.get_cls(data["prior"]["type"])
    return cls


def _build_parameters_from_dict(data, default_parameters):
    """Build a `~gammapy.modeling.Parameters` object from input dictionary and default parameter values."""
    par_data = []

    input_names = [_["name"] for _ in data]

    for par in default_parameters:
        par_dict = par.to_dict()
        try:
            index = input_names.index(par_dict["name"])
            par_dict.update(data[index])
        except ValueError:
            log.warning(
                f"Parameter '{par_dict['name']}' not defined in YAML file."
                f" Using default value: {par_dict['value']} {par_dict['unit']}"
            )
        par_data.append(par_dict)

    return Parameters.from_dict(par_data)


def _check_name_unique(model, names):
    """Check if a model is not duplicated."""
    if model.name in names:
        raise (
            ValueError(
                f"Model names must be unique. Models named '{model.name}' are duplicated."
            )
        )
    return


def _check_fov_background_models(models):
    """
    Check if a maximum of one `~gammapy.modeling.models.FoVBackgroundModel` is assigned to dataset
    and returns a dictionary mapping `dataset_name` to the background model name.

    Parameters
    ----------
    models : `~gammapy.modeling.models.Models`
        List of Models

    Returns
    -------
    bkg_model_mapping : dict
        Dictionary mapping dataset name to `~gammapy.modeling.models.FoVBackgroundModel` name.
    """
    from . import FoVBackgroundModel

    bkg_model_mapping = {}
    for model in models:
        if isinstance(model, FoVBackgroundModel):
            for n in model.datasets_names:
                if n not in bkg_model_mapping.keys():
                    bkg_model_mapping[n] = model.name
                else:
                    raise ValueError(
                        f"Only one FoVBackgroundModel per Dataset is permitted - already got one for {n}"
                    )
    return bkg_model_mapping


def _write_models(
    models,
    path,
    overwrite=False,
    full_output=False,
    overwrite_templates=False,
    write_covariance=True,
    checksum=False,
    extra_dict=None,
):
    """Write models to YAML file with additional information using an `extra_dict`."""

    base_path, _ = split(path)
    path = make_path(path)
    base_path = make_path(base_path)

    if path.exists() and not overwrite:
        raise IOError(f"File exists already: {path}")

    if (
        write_covariance
        and models.covariance is not None
        and len(models.parameters) != 0
    ):
        filecovar = path.stem + "_covariance.dat"
        kwargs = dict(format="ascii.fixed_width", delimiter="|", overwrite=overwrite)
        models.write_covariance(base_path / filecovar, **kwargs)
        models._covar_file = filecovar

    yaml_str = ""
    if extra_dict is not None:
        yaml_str += to_yaml(extra_dict)
    yaml_str += models.to_yaml(full_output, overwrite_templates)

    write_yaml(yaml_str, path, overwrite=overwrite, checksum=checksum)


class ModelBase:
    """Model base class."""

    _type = None

    def __init__(self, **kwargs):
        # Copy default parameters from the class to the instance
        default_parameters = self.default_parameters.copy()

        for key in kwargs.keys():
            if key != "covariance_data" and key not in default_parameters.names:
                raise NameError(f"Unknown Parameter name '{key}'")

        for par in default_parameters:
            value = kwargs.get(par.name, par)

            if not isinstance(value, Parameter):
                par.quantity = u.Quantity(value)
            else:
                par = value

            setattr(self, par.name, par)

        self._covariance = Covariance(self.parameters)

        covariance_data = kwargs.get("covariance_data", None)

        if covariance_data is not None:
            self.covariance = covariance_data

    def __getattribute__(self, name):
        value = object.__getattribute__(self, name)

        if isinstance(value, Parameter):
            return value.__get__(self, None)

        return value

    @property
    def type(self):
        return self._type

    def __init_subclass__(cls, **kwargs):
        # Add parameters list on the model sub-class (not instances)
        cls.default_parameters = Parameters(
            [_ for _ in cls.__dict__.values() if isinstance(_, Parameter)]
        )

    @classmethod
    def from_parameters(cls, parameters, **kwargs):
        """Create model from parameter list.

        Parameters
        ----------
        parameters : `Parameters`
            Parameters for init.

        Returns
        -------
        model : `Model`
            Model instance.
        """
        for par in parameters:
            kwargs[par.name] = par
        return cls(**kwargs)

    def _check_covariance(self):
        if not self.parameters == self._covariance.parameters:
            self._covariance = Covariance(self.parameters)

    @property
    def covariance(self):
        self._check_covariance()
        for par in self.parameters:
            pars = Parameters([par])
            error = np.nan_to_num(par.error**2, nan=1)
            covar = Covariance(pars, data=[[error]])
            self._covariance.set_subcovariance(covar)

        return self._covariance

    @covariance.setter
    def covariance(self, covariance):
        self._check_covariance()
        self._covariance.data = covariance

        for par in self.parameters:
            pars = Parameters([par])
            variance = self._covariance.get_subcovariance(pars).data
            par.error = np.sqrt(variance[0][0])

    @property
    def parameters(self):
        """Parameters as a `~gammapy.modeling.Parameters` object."""
        return Parameters(
            [getattr(self, name) for name in self.default_parameters.names]
        )

    @property
    def parameters_unique_names(self):
        return self.parameters.unique_parameters.names

    def copy(self, **kwargs):
        """Deep copy."""
        return copy.deepcopy(self)

    def to_dict(self, full_output=False):
        """Create dictionary for YAML serialisation."""
        tag = self.tag[0] if isinstance(self.tag, list) else self.tag
        params = self.parameters.to_dict()

        if not full_output:
            for par, par_default in zip(params, self.default_parameters):
                init = par_default.to_dict()
                for item in [
                    "min",
                    "max",
                    "error",
                    "interp",
                    "scale_method",
                    "scale_transform",
                ]:
                    default = init[item]

                    if par[item] == default or (
                        np.isnan(par[item]) and np.isnan(default)
                    ):
                        del par[item]

                if par["frozen"] == init["frozen"]:
                    del par["frozen"]

                if init["unit"] == "":
                    del par["unit"]

        data = {"type": tag, "parameters": params}

        if self.type is None:
            return data
        else:
            return {self.type: data}

    @classmethod
    def from_dict(cls, data, **kwargs):
        key0 = next(iter(data))

        if key0 in ["spatial", "temporal", "spectral"]:
            data = data[key0]

        if data["type"] not in cls.tag:
            raise ValueError(
                f"Invalid model type {data['type']} for class {cls.__name__}"
            )

        parameters = _build_parameters_from_dict(
            data["parameters"], cls.default_parameters
        )

        return cls.from_parameters(parameters, **kwargs)

    def __str__(self):
        string = f"{self.__class__.__name__}\n"
        if len(self.parameters) > 0:
            string += f"\n{self.parameters.to_table()}"
        return string

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def frozen(self):
        """Frozen status of a model, True if all parameters are frozen."""
        return np.all([p.frozen for p in self.parameters])

    def freeze(self):
        """Freeze all parameters."""
        self.parameters.freeze_all()

    def unfreeze(self):
        """Restore parameters frozen status to default."""
        for p, default in zip(self.parameters, self.default_parameters):
            p.frozen = default.frozen

    def reassign(self, datasets_names, new_datasets_names):
        """Reassign a model from one dataset to another.

        Parameters
        ----------
        datasets_names : str or list
            Name of the datasets where the model is currently defined.
        new_datasets_names : str or list
            Name of the datasets where the model should be defined instead.
            If multiple names are given the two list must have the save length,
            as the reassignment is element-wise.

        Returns
        -------
        model : `Model`
            Reassigned model.

        """
        model = self.copy(name=self.name)

        if not isinstance(datasets_names, list):
            datasets_names = [datasets_names]

        if not isinstance(new_datasets_names, list):
            new_datasets_names = [new_datasets_names]

        if isinstance(model.datasets_names, str):
            model.datasets_names = [model.datasets_names]

        if getattr(model, "datasets_names", None):
            for name, name_new in zip(datasets_names, new_datasets_names):
                model.datasets_names = [
                    _.replace(name, name_new) for _ in model.datasets_names
                ]

        return model


class Model:
    """Model class that contains only methods to create a model listed in the registries."""

    @staticmethod
    def create(tag, model_type=None, *args, **kwargs):
        """Create a model instance.

        Examples
        --------
        >>> from gammapy.modeling.models import Model
        >>> spectral_model = Model.create(
        ...            "pl-2", model_type="spectral", amplitude="1e-10 cm-2 s-1", index=3
        ...        )
        >>> type(spectral_model)
        <class 'gammapy.modeling.models.spectral.PowerLaw2SpectralModel'>
        """
        data = {"type": tag}
        if model_type is not None:
            data = {model_type: data}

        cls = _get_model_class_from_dict(data)
        return cls(*args, **kwargs)

    @staticmethod
    def from_dict(data):
        """Create a model instance from a dictionary."""
        cls = _get_model_class_from_dict(data)
        return cls.from_dict(data)


class DatasetModels(collections.abc.Sequence, CovarianceMixin):
    """Immutable models container.

    Parameters
    ----------
    models : `SkyModel`, list of `SkyModel` or `Models`
        Sky models.
    covariance_data : `~numpy.ndarray`
        Covariance data.
    """

    def __init__(self, models=None, covariance_data=None):
        if models is None:
            models = []

        if isinstance(models, (Models, DatasetModels)):
            if covariance_data is None and models.covariance is not None:
                covariance_data = models.covariance.data
            models = models._models
        elif isinstance(models, ModelBase):
            models = [models]
        elif not isinstance(models, list):
            raise TypeError(f"Invalid type: {models!r}")

        unique_names = []

        for model in models:
            _check_name_unique(model, names=unique_names)
            unique_names.append(model.name)

        self._background_models = _check_fov_background_models(models)

        self._models = models

        self._covar_file = None

        self._covariance = Covariance(self.parameters)

        # Set separately because this triggers the update mechanism on the sub-models
        if covariance_data is not None:
            self.covariance = covariance_data

    @property
    def parameters(self):
        """Parameters as a `~gammapy.modeling.Parameters` object."""
        return Parameters.from_stack([_.parameters for _ in self._models])

    @property
    def parameters_unique_names(self):
        """List of unique parameter names. Return formatted as model_name.par_type.par_name."""
        names = []
        for model in self._models:
            for par_name in model.parameters_unique_names:
                components = [model.name, par_name]
                name = ".".join(components)
                names.append(name)
        return names

    @property
    def names(self):
        """List of model names."""
        return [m.name for m in self._models]

    @property
    def background_models(self):
        """Dictionnary mapping of dataset names with their associated `~gammapy.modeling.models.FoVBackgroundModel` names."""
        return self._background_models

    @classmethod
    def read(cls, filename, checksum=False):
        """Read from YAML file.

        Parameters
        ----------
        filename : str
            input filename
        checksum : bool, optional
            Whether to perform checksum verification. Default is False.
        """
        yaml_str = make_path(filename).read_text()
        path, filename = split(filename)
        return cls.from_yaml(yaml_str, path=path, checksum=checksum)

    @classmethod
    def from_yaml(cls, yaml_str, path="", checksum=False):
        """Create from YAML string.

        Parameters
        ----------
        yaml_str : str
            yaml str
        path : str
            base path of model files
        checksum : bool, optional
            Whether to perform checksum verification. Default is False.


        """
        data = from_yaml(yaml_str, checksum=checksum)
        # TODO : for now metadata are not kept. Add proper metadata creation.
        data.pop("metadata", None)
        return cls.from_dict(data, path=path)

    @classmethod
    def from_dict(cls, data, path=""):
        """Create from dictionary."""
        from . import MODEL_REGISTRY

        path = make_path(path)

        models = []
        for component in data["components"]:
            model_cls = MODEL_REGISTRY.get_cls(component["type"])
            _recursive_dict_filename_update(component, path)
            model = model_cls.from_dict(component)
            _recursive_model_filename_update(model, path)
            models.append(model)
        models = cls(models)

        if "covariance" in data:
            filename = data["covariance"]
            if not (path / filename).exists():
                path, filename = split(filename)
            try:
                models.read_covariance(path, filename, format="ascii.fixed_width")
            except ValueError as exception:
                log.warning(
                    "Impossible to read the covariance correctly: \n"
                    f"{exception} \n "
                    "Covariance will be created from errors and non-diagonal terms ignored."
                )

        _set_models_link(models)

        return models

    def write(
        self,
        path,
        overwrite=False,
        full_output=False,
        overwrite_templates=False,
        write_covariance=True,
        checksum=False,
    ):
        """Write to YAML file.

        Parameters
        ----------
        path : `pathlib.Path` or str
            Path to write files.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        full_output : bool, optional
            Store full parameter output. Default is False.
        overwrite_templates : bool, optional
            Overwrite templates FITS files. Default is False.
        write_covariance : bool, optional
            Whether to save the covariance. Default is True.
        checksum : bool, optional
            When True adds a CHECKSUM entry to the file.
            Default is False.
        """
        path = make_path(path)
        for m in self:
            _recursive_model_filename_update(m, path)
        _write_models(
            self,
            path,
            overwrite,
            full_output,
            overwrite_templates,
            write_covariance,
            checksum,
        )

    def to_yaml(self, full_output=False, overwrite_templates=False):
        """Convert to YAML string.

        Parameters
        ----------
         full_output : bool, optional
            Store full parameter output. Default is False.
        overwrite_templates : bool, optional
            Overwrite templates FITS files. Default is False.
        """
        data = self.to_dict(full_output, overwrite_templates)
        return to_yaml(data)

    def update_link_label(self):
        """Update linked parameters labels used for serialisation and print."""
        self.parameters.update_link_label()

    def to_dict(self, full_output=False, overwrite_templates=False):
        """Convert to dictionary."""
        self.update_link_label()

        models_data = []
        for model in self._models:
            model_data = model.to_dict(full_output)
            models_data.append(model_data)
            if (
                hasattr(model, "spatial_model")
                and model.spatial_model is not None
                and "template" in model.spatial_model.tag
            ):
                model.spatial_model.write(overwrite=overwrite_templates)
            if model.tag == "TemplateNPredModel":
                model.write(overwrite=overwrite_templates)

        if self._covar_file is not None:
            return {
                "components": models_data,
                "covariance": str(self._covar_file),
            }
        else:
            return {"components": models_data}

    def to_parameters_table(self):
        """Convert model parameters to a `~astropy.table.Table`."""
        table = self.parameters.to_table()
        # Warning: splitting of parameters will break is source name has a "." in its name.
        model_name = [name.split(".")[0] for name in self.parameters_unique_names]
        table.add_column(model_name, name="model", index=0)
        return table

    def update_parameters_from_table(self, t):
        """Update models from a `~astropy.table.Table`."""
        parameters_dict = [dict(zip(t.colnames, row)) for row in t]
        for k, data in enumerate(parameters_dict):
            self.parameters[k].update_from_dict(data)

    def read_covariance(self, path, filename="_covariance.dat", **kwargs):
        """Read covariance data from file.

        Parameters
        ----------
        path : str or `Path`
            Base path.
        filename : str
            Filename.
        **kwargs : dict
            Keyword arguments passed to `~astropy.table.Table.read`.

        """
        path = make_path(path)
        filepath = str(path / filename)
        t = Table.read(filepath, **kwargs)
        t.remove_column("Parameters")
        arr = np.array(t)
        data = arr.view(float).reshape(arr.shape + (-1,))
        self.covariance = data
        self._covar_file = filename

    def write_covariance(self, filename, **kwargs):
        """Write covariance to file.

        Parameters
        ----------
        filename : str
            Filename.
        **kwargs : dict
            Keyword arguments passed to `~astropy.table.Table.write`.

        """
        names = self.parameters_unique_names
        table = Table()
        table["Parameters"] = names

        for idx, name in enumerate(names):
            values = self.covariance.data[idx]
            table[str(idx)] = values

        table.write(make_path(filename), **kwargs)

    def __str__(self):
        self.update_link_label()

        str_ = f"{self.__class__.__name__}\n\n"

        for idx, model in enumerate(self):
            str_ += f"Component {idx}: "
            str_ += str(model)

        return str_.expandtabs(tabsize=2)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def __add__(self, other):
        if isinstance(other, (Models, list)):
            return Models([*self, *other])
        elif isinstance(other, ModelBase):
            _check_name_unique(other, self.names)
            return Models([*self, other])
        else:
            raise TypeError(f"Invalid type: {other!r}")

    def __getitem__(self, key):
        if isinstance(key, np.ndarray) and key.dtype == bool:
            return self.__class__(list(np.array(self._models)[key]))
        else:
            return self._models[self.index(key)]

    def index(self, key):
        if isinstance(key, (int, slice)):
            return key
        elif isinstance(key, str):
            return self.names.index(key)
        elif isinstance(key, ModelBase):
            return self._models.index(key)
        else:
            raise TypeError(f"Invalid type: {type(key)!r}")

    def __len__(self):
        return len(self._models)

    def _ipython_key_completions_(self):
        return self.names

    def copy(self, copy_data=False):
        """Deep copy.

        Parameters
        ----------
        copy_data : bool
            Whether to copy data attached to template models.

        Returns
        -------
        models: `Models`
            Copied models.
        """
        self.update_link_label()
        models = []
        for model in self:
            model_copy = model.copy(name=model.name, copy_data=copy_data)
            models.append(model_copy)
        _set_models_link(models)

        return self.__class__(
            models=models, covariance_data=self.covariance.data.copy()
        )

    def _slice_by_energy(self, energy_min, energy_max, sum_over_energy_groups=False):
        """Copy models and slice TemplateNPredModel in energy range.

        Parameters
        ----------
        energy_min, energy_max : `~astropy.units.Quantity`
            Energy bounds of the slice
        sum_over_energy_groups : bool
            Whether to sum over the energy groups or not. Default is False.

        Returns
        -------
        models : `Models`
            Sliced models.
        """
        models_sliced = Models(self.copy())
        for k, m in enumerate(models_sliced):
            if m.tag == "TemplateNPredModel":
                m_sliced = m.slice_by_energy(energy_min, energy_max)
                if sum_over_energy_groups:
                    m_sliced.map = m_sliced.map.sum_over_axes(keepdims=True)
                models_sliced[k] = m_sliced
        return models_sliced

    def select(
        self,
        name_substring=None,
        datasets_names=None,
        tag=None,
        model_type=None,
        frozen=None,
    ):
        """Select models that meet all specified conditions.

        Parameters
        ----------
        name_substring : str, optional
            Substring contained in the model name. Default is None.
        datasets_names : str or list, optional
            Name of the dataset. Default is None.
        tag : str or list, optional
            Model tag. Default is None.
        model_type : {'None', 'spatial', 'spectral'}
           Type of model, used together with "tag", if the tag is not unique. Default is None.
        frozen : bool, optional
            If True, select models with all parameters frozen; if False, exclude them.  Default is None.

        Returns
        -------
        models : `DatasetModels`
            Selected models.
        """
        mask = self.selection_mask(
            name_substring, datasets_names, tag, model_type, frozen
        )
        return self[mask]

    def selection_mask(
        self,
        name_substring=None,
        datasets_names=None,
        tag=None,
        model_type=None,
        frozen=None,
    ):
        """Create a mask of models, that meet all specified conditions.

        Parameters
        ----------
        name_substring : str, optional
            Substring contained in the model name. Default is None.
        datasets_names : str or list of str, optional
            Name of the dataset. Default is None.
        tag : str or list of str, optional
            Model tag. Default is None.
        model_type : {'None', 'spatial', 'spectral'}
           Type of model, used together with "tag", if the tag is not unique. Default is None.
        frozen : bool, optional
            Select models with all parameters frozen if True, exclude them if False. Default is None.

        Returns
        -------
        mask : `numpy.array`
            Boolean mask, True for selected models.
        """
        selection = np.ones(len(self), dtype=bool)

        if tag and not isinstance(tag, list):
            tag = [tag]

        if datasets_names and not isinstance(datasets_names, list):
            datasets_names = [datasets_names]

        for idx, model in enumerate(self):
            if name_substring:
                selection[idx] &= name_substring in model.name

            if datasets_names:
                selection[idx] &= model.datasets_names is None or np.any(
                    [name in model.datasets_names for name in datasets_names]
                )

            if tag:
                if model_type is None:
                    sub_model = model
                else:
                    sub_model = getattr(model, f"{model_type}_model", None)

                if sub_model:
                    selection[idx] &= np.any([t in sub_model.tag for t in tag])
                else:
                    selection[idx] &= False

            if frozen is not None:
                if frozen:
                    selection[idx] &= model.frozen
                else:
                    selection[idx] &= ~model.frozen

        return np.array(selection, dtype=bool)

    def select_mask(self, mask, margin="0 deg", use_evaluation_region=True):
        """Check if sky models contribute within a mask map.

        Parameters
        ----------
        mask : `~gammapy.maps.WcsNDMap` of boolean type
            Map containing a boolean mask.
        margin : `~astropy.unit.Quantity`, optional
            Add a margin in degree to the source evaluation radius.
            Used to take into account PSF width. Default is "0 deg".
        use_evaluation_region : bool, optional
            Account for the extension of the model or not. Default is True.

        Returns
        -------
        models : `DatasetModels`
            Selected models contributing inside the region where mask==True.
        """
        models = []

        if not mask.geom.is_image:
            mask = mask.reduce_over_axes(func=np.logical_or)

        for model in self.select(tag="sky-model"):
            if use_evaluation_region:
                contributes = model.contributes(mask=mask, margin=margin)
            else:
                contributes = mask.get_by_coord(model.position, fill_value=0)

            if np.any(contributes):
                models.append(model)

        return self.__class__(models=models)

    def select_from_geom(self, geom, **kwargs):
        """Select models that fall inside a given geometry.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Geometry to select models from.
        **kwargs : dict
            Keyword arguments passed to `~gammapy.modeling.models.DatasetModels.select_mask`.

        Returns
        -------
        models : `DatasetModels`
            Selected models.
        """
        mask = Map.from_geom(geom=geom, data=True, dtype=bool)
        return self.select_mask(mask=mask, **kwargs)

    def select_region(self, regions, wcs=None):
        """Select sky models with center position contained within a given region.

        Parameters
        ----------
        regions : str, `~regions.Region` or list of `~regions.Region`
            Region or list of regions (pixel or sky regions accepted).
            A region can be defined as a string ind DS9 format as well.
            See http://ds9.si.edu/doc/ref/region.html for details.
        wcs : `~astropy.wcs.WCS`, optional
            World coordinate system transformation. Default is None.

        Returns
        -------
        models : `DatasetModels`
            Selected models.
        """
        geom = RegionGeom.from_regions(regions, wcs=wcs)

        models = []

        for model in self.select(tag="sky-model"):
            if geom.contains(model.position):
                models.append(model)

        return self.__class__(models=models)

    def restore_status(self, restore_values=True):
        """Context manager to restore status.

        A copy of the values is made on enter,
        and those values are restored on exit.

        Parameters
        ----------
        restore_values : bool, optional
            Restore values if True, otherwise restore only frozen status and covariance matrix.
            Default is True.

        """
        return restore_models_status(self, restore_values)

    def set_parameters_bounds(
        self, tag, model_type, parameters_names=None, min=None, max=None, value=None
    ):
        """Set bounds for the selected models types and parameters names.

        Parameters
        ----------
        tag : str or list
            Tag of the models.
        model_type : {"spatial", "spectral", "temporal"}
            Type of model.
        parameters_names : str or list, optional
            Parameters names. Default is None.
        min : float, optional
            Minimum value. Default is None.
        max : float, optional
            Maximum value. Default is None.
        value : float, optional
            Initial value. Default is None.
        """
        models = self.select(tag=tag, model_type=model_type)
        parameters = models.parameters.select(name=parameters_names, type=model_type)
        n = len(parameters)

        if min is not None:
            parameters.min = np.ones(n) * min
        if max is not None:
            parameters.max = np.ones(n) * max
        if value is not None:
            parameters.value = np.ones(n) * value

    def freeze(self, model_type=None):
        """Freeze parameters depending on model type.

        Parameters
        ----------
        model_type : {None, "spatial", "spectral"}
           Freeze all parameters or only spatial or only spectral.
           Default is None.
        """
        for m in self:
            m.freeze(model_type)

    def unfreeze(self, model_type=None):
        """Restore parameters frozen status to default depending on model type.

        Parameters
        ----------
        model_type : {None, "spatial", "spectral"}
           Restore frozen status to default for all parameters or only spatial or only spectral.
           Default is None.
        """
        for m in self:
            m.unfreeze(model_type)

    @property
    def frozen(self):
        """Boolean mask, True if all parameters of a given model are frozen."""
        return np.all([m.frozen for m in self])

    def reassign(self, dataset_name, new_dataset_name):
        """Reassign a model from one dataset to another.

        Parameters
        ----------
        dataset_name : str or list
            Name of the datasets where the model is currently defined.
        new_dataset_name : str or list
            Name of the datasets where the model should be defined instead.
            If multiple names are given the two list must have the save length,
            as the reassignment is element-wise.
        """
        models = [m.reassign(dataset_name, new_dataset_name) for m in self]
        return self.__class__(models)

    def to_template_sky_model(self, geom, spectral_model=None, name=None):
        """Merge a list of models into a single `~gammapy.modeling.models.SkyModel`.

        Parameters
        ----------
        geom : `Geom`
            Map geometry of the result template model.
        spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
            One of the NormSpectralModel. Default is None.
        name : str, optional
            Name of the new model. Default is None.

        Returns
        -------
        model : `SkyModel`
            Template sky model.
        """
        from . import PowerLawNormSpectralModel, SkyModel, TemplateSpatialModel

        unit = u.Unit("1 / (cm2 s sr TeV)")
        map_ = Map.from_geom(geom, unit=unit)

        for m in self:
            map_ += m.evaluate_geom(geom).to(unit)

        spatial_model = TemplateSpatialModel(map_, normalize=False)

        if spectral_model is None:
            spectral_model = PowerLawNormSpectralModel()

        return SkyModel(
            spectral_model=spectral_model, spatial_model=spatial_model, name=name
        )

    def to_template_spectral_model(self, geom, mask=None):
        """Merge a list of models into a single `~gammapy.modeling.models.TemplateSpectralModel`.

        For each model the spatial component is integrated over the given geometry where the mask is true
        and multiplied by the spectral component value in each energy bin.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Map geometry on which the template model is computed.
        mask :  `~gammapy.maps.Map` with bool dtype.
            Evaluate the model only where the mask is True.

        Returns
        -------
        model : `~gammapy.modeling.models.TemplateSpectralModel`
            Template spectral model.
        """
        from . import TemplateSpectralModel

        energy = geom.axes[0].center
        if mask is None:
            mask = Map.from_geom(geom, data=True)
        elif mask.geom != geom:
            mask = mask.interp_to_geom(geom, method="nearest")
        values = 0
        for m in self:
            dnde = m.spectral_model(energy)
            spatial_integ = m.spatial_model.integrate_geom(geom)
            masked_integ = spatial_integ.data * np.isfinite(spatial_integ.data) * mask
            spatial_integ.data[~np.isfinite(spatial_integ.data)] = np.nan
            dnde *= masked_integ.sum(axis=(1, 2)) * spatial_integ.unit
            values += dnde
        return TemplateSpectralModel(energy, values)

    @property
    def positions(self):
        """Positions of the models as a `~astropy.coordinates.SkyCoord`."""
        positions = []

        for model in self.select(tag="sky-model"):
            if model.position:
                positions.append(model.position.icrs)
            else:
                log.warning(
                    f"Skipping model {model.name} - no spatial component present"
                )

        return SkyCoord(positions)

    def to_regions(self, size_factor=None, **kwargs):
        """Return a list of the regions for the spatial models.

        Parameters
        ----------
        size_factor : float, optional
            Factor applied to the size of the models.
            If not specified, the defaults for the models will be used.
        kwargs : dict, optional
            Keyword arguments passed to ``model.spatial_model.to_region``.

        Returns
        -------
        regions : list of `~regions.SkyRegion`
            Regions.
        """
        regions = []
        if size_factor:
            kwargs["size_factor"] = size_factor
        for model in self.select(tag="sky-model"):
            try:
                region = model.spatial_model.to_region(**kwargs)
                regions.append(region)
            except AttributeError:
                log.warning(
                    f"Skipping model {model.name} - no spatial component present"
                )
        return regions

    @property
    def wcs_geom(self):
        """Minimum WCS geom in which all the models are contained."""
        regions = self.to_regions()
        try:
            return RegionGeom.from_regions(regions).to_wcs_geom()
        except IndexError:
            log.error("No spatial component in any model. Geom not defined")

    def plot_regions(
        self, ax=None, kwargs_point=None, path_effect=None, size_factor=None, **kwargs
    ):
        """Plot extent of the spatial models on a given WCS axis.

        Parameters
        ----------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
            Axes to plot on. If no axes are given, an all-sky WCS
            is chosen using a CAR projection. Default is None.
        kwargs_point : dict, optional
            Keyword arguments passed to `~matplotlib.lines.Line2D` for plotting
            of point sources. Default is None.
        path_effect : `~matplotlib.patheffects.PathEffect`, optional
            Path effect applied to artists and lines. Default is None.
        size_factor : float, optional
            Factor applied to the size of the model
            If not specified, the defaults for the models will be used.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.artists.Artist`.

        Returns
        -------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes.

        Examples
        --------
        >>> from gammapy.datasets import MapDataset
        >>> from gammapy.catalog import SourceCatalog3FHL
        >>> fermi_dataset = MapDataset.read(
        ...    "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc.fits.gz", name="fermi_dataset")
        >>> catalog = SourceCatalog3FHL()
        >>> models = catalog.to_models().select_from_geom(fermi_dataset.geoms["geom"])
        >>> ax = fermi_dataset.excess.sum_over_axes().smooth("0.2 deg").plot(add_cbar=True, cmap="Blues")
        >>> ax = models.plot_regions(ax=ax, linewidth=1,  color="red",
        ...    kwargs_point={"marker":"o", "markersize":5, "color":"red"}
        ...            )
        """
        regions = self.to_regions(size_factor=size_factor)

        geom = RegionGeom.from_regions(regions=regions)
        return geom.plot_region(
            ax=ax, kwargs_point=kwargs_point, path_effect=path_effect, **kwargs
        )

    def plot_positions(self, ax=None, **kwargs):
        """Plot the centers of the spatial models on a given WCS axis.

        Parameters
        ----------
        ax : `~astropy.visualization.WCSAxes`, optional
            Axes to plot on. If no axes are given, an all-sky WCS
            is chosen using a CAR projection. Default is None.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.scatter`.


        Returns
        -------
        ax : `~astropy.visualization.WcsAxes`
            WCS axes.

        Examples
        --------

        >>> from gammapy.datasets import MapDataset
        >>> from gammapy.catalog import SourceCatalog3FHL
        >>> fermi_dataset = MapDataset.read(
        ...        "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc.fits.gz", name="fermi_dataset"
        ...        )
        >>> catalog = SourceCatalog3FHL()
        >>> models = catalog.to_models().select_from_geom(fermi_dataset.geoms["geom"])
        >>> ax = fermi_dataset.excess.sum_over_axes().smooth("0.2 deg").plot(add_cbar=True, cmap="Blues")
        >>> ax = models.plot_positions(ax=ax, color="red", marker="+", linewidth=1)
        """
        from astropy.visualization.wcsaxes import WCSAxes

        if ax is None or not isinstance(ax, WCSAxes):
            ax = Map.from_geom(self.wcs_geom).plot()

        kwargs.setdefault("marker", "*")
        kwargs.setdefault("color", "tab:blue")
        path_effects = kwargs.get("path_effects", None)

        xp, yp = self.positions.to_pixel(ax.wcs)
        p = ax.scatter(xp, yp, **kwargs)

        if path_effects:
            plt.setp(p, path_effects=path_effects)

        return ax


class Models(DatasetModels, collections.abc.MutableSequence):
    """Sky model collection.

    Parameters
    ----------
    models : `SkyModel`, list of `SkyModel` or `Models`
        Sky models.
    """

    def __delitem__(self, key):
        del self._models[self.index(key)]
        self._background_models = _check_fov_background_models(self._models)

    def __setitem__(self, key, model):
        from gammapy.modeling.models import (
            FoVBackgroundModel,
            SkyModel,
            TemplateNPredModel,
        )

        if isinstance(model, (SkyModel, FoVBackgroundModel, TemplateNPredModel)):
            ind = self.index(key)
            other_names = [_ for _ in self.names if _ != self._models[ind].name]
            _check_name_unique(model, other_names)
            self._models[ind] = model
        else:
            raise TypeError(f"Invalid type: {model!r}")

    def insert(self, idx, model):
        _check_name_unique(model, self.names)
        self._models.insert(idx, model)
        self._background_models = _check_fov_background_models(self._models)

    def set_prior(self, parameters, priors):
        for parameter, prior in zip(parameters, priors):
            parameter.prior = prior


class restore_models_status:
    def __init__(self, models, restore_values=True):
        self.restore_values = restore_values
        self.models = models
        self.values = [_.value for _ in models.parameters]
        self.frozen = [_.frozen for _ in models.parameters]
        self.covariance_data = models.covariance.data

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        for value, par, frozen in zip(self.values, self.models.parameters, self.frozen):
            if self.restore_values:
                par.value = value
            par.frozen = frozen
        self.models.covariance = self.covariance_data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utility functions to deal with arrays and quantities."""
import numpy as np
import scipy.ndimage
import scipy.signal
from astropy.convolution import Gaussian2DKernel

__all__ = [
    "array_stats_str",
    "round_up_to_even",
    "round_up_to_odd",
    "shape_2N",
    "shape_divisible_by",
    "symmetric_crop_pad_width",
]


def is_power2(n):
    """Check if an integer is a power of 2."""
    return (n > 0) & ((n & (n - 1)) == 0)


def array_stats_str(x, label=""):
    """Make a string summarising some stats for an array.

    Parameters
    ----------
    x : array-like
        Array.
    label : str, optional
        Label.

    Returns
    -------
    stats_str : str
        String with array stats.
    """
    x = np.asanyarray(x)

    ss = ""
    if label:
        ss += f"{label:15s}: "

    min = x.min()
    max = x.max()
    size = x.size

    fmt = "size = {size:5d}, min = {min:6.3f}, max = {max:6.3f}\n"
    ss += fmt.format(**locals())

    return ss


def shape_2N(shape, N=3):
    """
    Round a given shape to values that are divisible by 2^N.

    Parameters
    ----------
    shape : tuple
        Input shape.
    N : int, optional
        Exponent of two. Default is 3.

    Returns
    -------
    new_shape : tuple
        New shape extended to integers divisible by 2^N.
    """
    shape = np.array(shape)
    new_shape = shape + (2**N - np.mod(shape, 2**N))
    return tuple(new_shape)


def shape_divisible_by(shape, factor):
    """
    Round a given shape to values that are divisible by factor.

    Parameters
    ----------
    shape : tuple
        Input shape.
    factor : int
        Divisor.

    Returns
    -------
    new_shape : tuple
        New shape extended to integers divisible by factor.
    """
    shape = np.array(shape)
    new_shape = shape + (shape % factor)
    return tuple(new_shape)


def round_up_to_odd(f):
    """Round float to odd integer.

    Parameters
    ----------
    f : float
        Float value.

    Returns
    -------
    int : int
        Odd integer.
    """
    return (np.ceil(f) // 2 * 2 + 1).astype(int)


def round_up_to_even(f):
    """Round float to even integer.

    Parameters
    ----------
    f : float
        Float value.

    Returns
    -------
    int : int
        Odd integer.
    """
    return (np.ceil(f + 1) // 2 * 2).astype(int)


def symmetric_crop_pad_width(shape, new_shape):
    """
    Compute symmetric crop or pad width.

    To obtain a new shape from a given old shape of an array.

    Parameters
    ----------
    shape : tuple
        Old shape.
    new_shape : tuple or str
        New shape.
    """
    xdiff = abs(shape[1] - new_shape[1])
    ydiff = abs(shape[0] - new_shape[0])

    if (np.array([xdiff, ydiff]) % 2).any():
        raise ValueError(
            "For symmetric crop / pad width, difference to new shape "
            "must be even in all axes."
        )

    ywidth = (ydiff // 2, ydiff // 2)
    xwidth = (xdiff // 2, xdiff // 2)
    return ywidth, xwidth


def _fftconvolve_wrap(kernel, data):
    # wrap gaussian filter as a special case, because the gain in
    # performance is factor ~100
    if isinstance(kernel, Gaussian2DKernel):
        width = kernel.model.x_stddev.value
        norm = kernel.array.sum()
        return norm * scipy.ndimage.gaussian_filter(data, width)
    else:
        return scipy.signal.fftconvolve(
            data.astype(np.float32), kernel.array, mode="same"
        )


def scale_cube(data, kernels):
    """
    Compute scale space cube.

    Compute scale space cube by convolving the data with a set of kernels and
    stack the resulting images along the third axis.

    Parameters
    ----------
    data : `~numpy.ndarray`
        Input data.
    kernels : list of `~astropy.convolution.Kernel`
        List of convolution kernels.

    Returns
    -------
    cube : `~numpy.ndarray`
        Array of the shape (len(kernels), data.shape).
    """
    return np.dstack([_fftconvolve_wrap(kernel, data) for kernel in kernels])



"""
Make a theta-square plot
========================

This tutorial explains how to make such a plot, that is the distribution
of event counts as a function of the squared angular distance, to a test
position.

"""


######################################################################
# Setup
# -----
#

# %matplotlib inline
import matplotlib.pyplot as plt
from astropy.coordinates import SkyCoord
from astropy import units as u
from gammapy.data import DataStore
from gammapy.maps import MapAxis
from gammapy.makers.utils import make_theta_squared_table
from gammapy.visualization import plot_theta_squared_table


######################################################################
# Get some data
# -------------
#
# Some data taken on the Crab by H.E.S.S. are used.
#

data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
observations = data_store.get_observations([23523, 23526])


######################################################################
# Define a test position
# -----------------------
#
# Here we define the position of Crab
#

position = SkyCoord(83.6324, 22.0174, unit="deg")
print(position)


######################################################################
# Creation of the theta2 plot
# ---------------------------
#

theta2_axis = MapAxis.from_bounds(0, 0.2, nbin=20, interp="lin", unit="deg2")
theta2_table = make_theta_squared_table(
    observations=observations,
    position=position,
    theta_squared_axis=theta2_axis,
)

plt.figure(figsize=(10, 5))
plot_theta_squared_table(theta2_table)
plt.show()


######################################################################
# Making a theta2 plot for a given energy range
# ---------------------------------------------
#
# with the function `~gammapy.makers.utils.make_theta_squared_table`, one can
# also select a fixed energy range.
#

theta2_table_en = make_theta_squared_table(
    observations=observations,
    position=position,
    theta_squared_axis=theta2_axis,
    energy_edges=[1.2, 11] * u.TeV,
)
plt.figure(figsize=(10, 5))
plot_theta_squared_table(theta2_table_en)
plt.show()


######################################################################
# Statistical significance of a detection
# ---------------------------------------
#
# To get the significance of a signal, the usual method consists of using the reflected background method
# (see the maker tutorial: :doc:`/user-guide/makers/reflected`) to compute the WStat statistics
# (see :ref:`wstat`, :doc:`/user-guide/stats/fit_statistics`). This is the well-known method of [LiMa1983]_
# using ON and OFF regions.
#
# The following tutorials show how to get an excess significance:
#
# -  :doc:`/tutorials/analysis-1d/spectral_analysis`
# -  :doc:`/tutorials/analysis-1d/extended_source_spectral_analysis`
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .kernel import PSFKernel
from .map import PSFMap, RecoPSFMap
from .parametric import EnergyDependentMultiGaussPSF, ParametricPSF, PSFKing
from .table import PSF3D

__all__ = [
    "EnergyDependentMultiGaussPSF",
    "ParametricPSF",
    "PSF3D",
    "PSFKernel",
    "PSFKing",
    "PSFMap",
    "RecoPSFMap",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import logging
from astropy.table import Column, Table
from gammapy.maps import Map
from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
from gammapy.stats import WStatCountsStatistic
from ..core import Estimator
from ..utils import apply_threshold_sensitivity

log = logging.getLogger(__name__)

__all__ = ["SensitivityEstimator"]


class SensitivityEstimator(Estimator):
    """Estimate sensitivity.

    This class allows to determine for each reconstructed energy bin the flux
    associated to the number of gamma-ray events for which the significance is
    ``n_sigma``, and being larger than ``gamma_min`` and ``bkg_sys`` percent
    larger than the number of background events in the ON region.


    Parameters
    ----------
    spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
        Spectral model assumption. Default is power-law with spectral index of 2.
    n_sigma : float, optional
        Minimum significance. Default is 5.
    gamma_min : float, optional
        Minimum number of gamma-rays. Default is 10.
    bkg_syst_fraction : float, optional
        Fraction of background counts above which the number of gamma-rays is. Default is 0.05.

    Examples
    --------
    For a usage example see :doc:`/tutorials/analysis-1d/cta_sensitivity` tutorial.

    """

    tag = "SensitivityEstimator"

    def __init__(
        self,
        spectral_model=None,
        n_sigma=5.0,
        gamma_min=10,
        bkg_syst_fraction=0.05,
    ):
        if spectral_model is None:
            spectral_model = PowerLawSpectralModel(
                index=2, amplitude="1 cm-2 s-1 TeV-1"
            )

        self.spectral_model = spectral_model
        self.n_sigma = n_sigma
        self.gamma_min = gamma_min
        self.bkg_syst_fraction = bkg_syst_fraction

    def estimate_min_excess(self, dataset):
        """Estimate minimum excess to reach the given significance.

        Parameters
        ----------
        dataset : `~gammapy.datasets.SpectrumDataset`
            Spectrum dataset.

        Returns
        -------
        excess : `~gammapy.maps.RegionNDMap`
            Minimal excess.
        """
        n_off = dataset.counts_off.data

        stat = WStatCountsStatistic(
            n_on=dataset.alpha.data * n_off, n_off=n_off, alpha=dataset.alpha.data
        )
        excess_counts = stat.n_sig_matching_significance(self.n_sigma)

        excess_counts = apply_threshold_sensitivity(
            dataset.background.data,
            excess_counts,
            self.gamma_min,
            self.bkg_syst_fraction,
        )

        excess = Map.from_geom(geom=dataset._geom, data=excess_counts)
        return excess

    def estimate_min_e2dnde(self, excess, dataset):
        """Estimate e2dnde from a given minimum excess.

        Parameters
        ----------
        excess : `~gammapy.maps.RegionNDMap`
            Minimal excess.
        dataset : `~gammapy.datasets.SpectrumDataset`
            Spectrum dataset.

        Returns
        -------
        e2dnde : `~astropy.units.Quantity`
            Minimal differential flux.
        """
        energy = dataset._geom.axes["energy"].center

        dataset.models = SkyModel(spectral_model=self.spectral_model)
        npred = dataset.npred_signal()

        phi_0 = excess / npred

        dnde_model = self.spectral_model(energy=energy)
        e2dnde = phi_0.data[:, 0, 0] * dnde_model * energy**2
        return e2dnde.to("erg / (cm2 s)")

    def _get_criterion(self, excess, bkg):
        is_gamma_limited = excess == self.gamma_min
        is_bkg_syst_limited = excess == bkg * self.bkg_syst_fraction
        criterion = np.chararray(excess.shape, itemsize=12)
        criterion[is_gamma_limited] = "gamma"
        criterion[is_bkg_syst_limited] = "bkg"
        criterion[~np.logical_or(is_gamma_limited, is_bkg_syst_limited)] = (
            "significance"
        )
        return criterion

    def run(self, dataset):
        """Run the sensitivity estimation.

        Parameters
        ----------
        dataset : `~gammapy.datasets.SpectrumDatasetOnOff`
            Dataset to compute sensitivity for.

        Returns
        -------
        sensitivity : `~astropy.table.Table`
            Sensitivity table. Containing the following columns:

                * e_ref : energy center
                * e_min : minimum energy values
                * e_max : maximum energy values
                * e2dnde : minimal differential flux
                * excess : number of excess counts in the bin
                * background : number of background counts in the bin
                * criterion : sensitivity-limiting criterion

        """
        energy = dataset._geom.axes["energy"].center

        if np.any(self.spectral_model(energy).value < 0.0):
            log.warning(
                "Spectral model predicts negative flux. Results of estimator should be interpreted with caution"
            )

        excess = self.estimate_min_excess(dataset)
        e2dnde = self.estimate_min_e2dnde(excess, dataset)
        criterion = self._get_criterion(
            excess.data.squeeze(), dataset.background.data.squeeze()
        )

        return Table(
            [
                Column(
                    data=energy,
                    name="e_ref",
                    format="5g",
                    description="Energy center",
                ),
                Column(
                    data=dataset._geom.axes["energy"].edges_min,
                    name="e_min",
                    format="5g",
                    description="Energy edge low",
                ),
                Column(
                    data=dataset._geom.axes["energy"].edges_max,
                    name="e_max",
                    format="5g",
                    description="Energy edge high",
                ),
                Column(
                    data=e2dnde,
                    name="e2dnde",
                    format="5g",
                    description="Energy squared times differential flux",
                ),
                Column(
                    data=np.atleast_1d(excess.data.squeeze()),
                    name="excess",
                    format="5g",
                    description="Number of excess counts in the bin",
                ),
                Column(
                    data=np.atleast_1d(dataset.background.data.squeeze()),
                    name="background",
                    format="5g",
                    description="Number of background counts in the bin",
                ),
                Column(
                    data=np.atleast_1d(criterion),
                    name="criterion",
                    description="Sensitivity-limiting criterion",
                ),
            ]
        )



"""Plot the energy dispersion at a given offset."""

from astropy.coordinates import Angle
import matplotlib.pyplot as plt
from gammapy.irf import EnergyDispersion2D

filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
edisp = EnergyDispersion2D.read(filename, hdu="EDISP")
# offset at which we want to examine the energy dispersion
offset = Angle("0.5 deg")
edisp_kernel = edisp.to_edisp_kernel(offset=offset)
edisp_kernel.peek()
plt.show()



import pytest
from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord
from pydantic import ValidationError
from gammapy.datasets import MapDatasetMetaData
from gammapy.utils.metadata import ObsInfoMetaData, PointingInfoMetaData


def test_meta_default():
    meta = MapDatasetMetaData()
    assert meta.creation.creator.split()[0] == "Gammapy"
    assert meta.obs_info is None


def test_mapdataset_metadata():
    position = SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
    obs_info_input = {
        "telescope": "cta-north",
        "instrument": "lst",
        "observation_mode": "wobble",
        "obs_id": 112,
    }
    input = {
        "obs_info": ObsInfoMetaData(**obs_info_input),
        "pointing": PointingInfoMetaData(radec_mean=position),
        "optional": dict(test=0.5, other=True),
    }
    meta = MapDatasetMetaData(**input)

    assert meta.obs_info.telescope == "cta-north"
    assert meta.obs_info.instrument == "lst"
    assert meta.obs_info.observation_mode == "wobble"
    assert_allclose(meta.pointing.radec_mean.dec.value, 22.0147)
    assert_allclose(meta.pointing.radec_mean.ra.deg, 83.6287)
    assert meta.obs_info.obs_id == 112
    assert meta.optional["other"] is True
    assert meta.creation.creator.split()[0] == "Gammapy"
    assert meta.event_type is None

    with pytest.raises(ValidationError):
        meta.pointing = 2.0

    input_bad = input.copy()
    input_bad["bad"] = position

    with pytest.raises(ValueError):
        MapDatasetMetaData(**input_bad)


def test_mapdataset_metadata_lists():
    obs_info_input1 = {
        "telescope": "cta-north",
        "instrument": "lst",
        "observation_mode": "wobble",
        "obs_id": 111,
    }
    obs_info_input2 = {
        "telescope": "cta-north",
        "instrument": "lst",
        "observation_mode": "wobble",
        "obs_id": 112,
    }
    input = {
        "obs_info": [
            ObsInfoMetaData(**obs_info_input1),
            ObsInfoMetaData(**obs_info_input2),
        ],
        "pointing": [
            PointingInfoMetaData(
                radec_mean=SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
            ),
            PointingInfoMetaData(
                radec_mean=SkyCoord(83.1287, 22.5147, unit="deg", frame="icrs")
            ),
        ],
    }
    meta = MapDatasetMetaData(**input)
    assert meta.obs_info[0].telescope == "cta-north"
    assert meta.obs_info[0].instrument == "lst"
    assert meta.obs_info[0].observation_mode == "wobble"
    assert_allclose(meta.pointing[0].radec_mean.dec.value, 22.0147)
    assert_allclose(meta.pointing[1].radec_mean.ra.deg, 83.1287)
    assert meta.obs_info[0].obs_id == 111
    assert meta.obs_info[1].obs_id == 112
    assert meta.optional is None
    assert meta.event_type is None


def test_mapdataset_metadata_stack():
    input1 = {
        "obs_info": ObsInfoMetaData(**{"obs_id": 111}),
        "pointing": PointingInfoMetaData(
            radec_mean=SkyCoord(83.6287, 22.5147, unit="deg", frame="icrs")
        ),
        "optional": dict(test=0.5, other=True),
    }

    input2 = {
        "obs_info": ObsInfoMetaData(**{"instrument": "H.E.S.S.", "obs_id": 112}),
        "pointing": PointingInfoMetaData(
            radec_mean=SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
        ),
        "optional": dict(test=0.1, other=False),
    }

    meta1 = MapDatasetMetaData(**input1)
    meta2 = MapDatasetMetaData(**input2)

    meta = meta1.stack(meta2)
    assert meta.creation.creator.split()[0] == "Gammapy"
    assert meta.obs_info is None



import json
from pathlib import Path
from typing import Annotated
from astropy import units as u
from astropy.coordinates import AltAz, Angle, EarthLocation, SkyCoord
from astropy.time import Time
from pydantic import PlainSerializer
from pydantic.functional_validators import AfterValidator, BeforeValidator
from .observers import observatory_locations
from .scripts import make_path

__all__ = [
    "AngleType",
    "EnergyType",
    "QuantityType",
    "TimeType",
    "PathType",
    "EarthLocationType",
    "SkyCoordType",
]


# TODO: replace by QuantityType and pydantic TypeAdapter
class JsonQuantityEncoder(json.JSONEncoder):
    """Support for quantities that JSON default encoder."""

    def default(self, obj):
        if isinstance(obj, u.Quantity):
            return obj.to_string()

        return json.JSONEncoder.default(self, obj)


# TODO: replace by QuantityType and pydantic TypeAdapter
class JsonQuantityDecoder(json.JSONDecoder):
    """Support for quantities that JSON default encoder."""

    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(data):
        for key, value in data.items():
            try:
                data[key] = u.Quantity(value)
            except TypeError:
                continue
        return data


def json_encode_earth_location(v):
    """JSON encoder for `~astropy.coordinates.EarthLocation`."""
    return (
        f"lon: {v.lon.value} {v.lon.unit}, "
        f"lat : {v.lat.value} {v.lat.unit}, "
        f"height : {v.height.value} {v.height.unit}"
    )


def json_encode_sky_coord(v):
    """JSON encoder for `~astropy.coordinates.SkyCoord`."""
    return f"lon: {v.spherical.lon.value} {v.spherical.lon.unit}, lat: {v.spherical.lat.value} {v.spherical.lat.unit}, frame: {v.frame.name} "


def json_encode_time(v):
    """JSON encoder for `~astropy.time.Time`."""
    if v.isscalar:
        return v.value

    return v.value.tolist()


def validate_angle(v):
    """Validator for `~astropy.coordinates.Angle`."""
    return Angle(v)


def validate_scalar(v):
    """Validator for scalar values."""
    if not v.isscalar:
        raise ValueError(f"A scalar value is required: {v!r}")

    return v


def validate_energy(v):
    """Validator for `~astropy.units.Quantity` with unit "energy"."""
    v = u.Quantity(v)

    if v.unit.physical_type != "energy":
        raise ValueError(f"Invalid unit for energy: {v.unit!r}")

    return v


def validate_quantity(v):
    """Validator for `~astropy.units.Quantity`."""
    return u.Quantity(v)


def validate_time(v):
    """Validator for `~astropy.time.Time`."""
    return Time(v)


def validate_earth_location(v):
    """Validator for `~astropy.coordinates.EarthLocation`."""
    if isinstance(v, EarthLocation):
        return v

    if isinstance(v, str) and v in observatory_locations:
        return observatory_locations[v]

    try:
        return EarthLocation(v)
    except TypeError:
        raise ValueError(f"Invalid EarthLocation: {v!r}")


def validate_sky_coord(v):
    """Validator for `~astropy.coordinates.SkyCoord`."""
    return SkyCoord(v)


def validate_sky_coord_icrs(v):
    """Validator for `~astropy.coordinates.SkyCoord` in icrs."""
    try:
        return SkyCoord(v).icrs
    except AttributeError:
        raise ValueError(f"Cannot convert '{v!r}' to icrs")


def validate_altaz_coord(v):
    """Validator for `~astropy.coordinates.AltAz`."""
    if isinstance(v, AltAz):
        return SkyCoord(v)

    return SkyCoord(v).altaz


SERIALIZE_KWARGS = {
    "when_used": "json-unless-none",
    "return_type": str,
}

scalar_validator = AfterValidator(validate_scalar)


AngleType = Annotated[
    Angle,
    PlainSerializer(lambda v: f"{v.value} {v.unit}", **SERIALIZE_KWARGS),
    BeforeValidator(validate_angle),
    scalar_validator,
]

EnergyType = Annotated[
    u.Quantity,
    PlainSerializer(lambda v: f"{v.value} {v.unit}", **SERIALIZE_KWARGS),
    BeforeValidator(validate_energy),
    scalar_validator,
]

QuantityType = Annotated[
    u.Quantity,
    PlainSerializer(lambda v: f"{v.value} {v.unit}", **SERIALIZE_KWARGS),
    BeforeValidator(validate_quantity),
    scalar_validator,
]

TimeType = Annotated[
    Time,
    PlainSerializer(json_encode_time, when_used="json-unless-none"),
    BeforeValidator(validate_time),
]


EarthLocationType = Annotated[
    EarthLocation,
    PlainSerializer(json_encode_earth_location, **SERIALIZE_KWARGS),
    BeforeValidator(validate_earth_location),
    scalar_validator,
]

SkyCoordType = Annotated[
    SkyCoord,
    PlainSerializer(json_encode_sky_coord, **SERIALIZE_KWARGS),
    BeforeValidator(validate_sky_coord),
    scalar_validator,
]

ICRSSkyCoordType = Annotated[
    SkyCoord,
    PlainSerializer(json_encode_sky_coord, **SERIALIZE_KWARGS),
    BeforeValidator(validate_sky_coord_icrs),
    scalar_validator,
]

AltAzSkyCoordType = Annotated[
    SkyCoord,
    PlainSerializer(json_encode_sky_coord, **SERIALIZE_KWARGS),
    BeforeValidator(validate_altaz_coord),
    scalar_validator,
]

PathType = Annotated[
    Path,
    PlainSerializer(lambda p: str(p), **SERIALIZE_KWARGS),
    BeforeValidator(make_path),
]



r"""
.. _logparabola-spectral-model:

Log parabola spectral model
===========================

This model parametrises a log parabola spectrum.

It is defined by the following equation:

.. math::
        \phi(E) = \phi_0 \left( \frac{E}{E_0} \right) ^ {
          - \alpha - \beta \log{ \left( \frac{E}{E_0} \right) }
        }

Note that :math:`log` refers to the natural logarithm. This is consistent
with the `Fermi Science Tools
<https://fermi.gsfc.nasa.gov/ssc/data/analysis/scitools/source_models.html>`_
and `ctools
<http://cta.irap.omp.eu/ctools-devel/users/user_manual/getting_started/models.html#log-parabola>`_.
The `Sherpa <http://cxc.harvard.edu/sherpa/ahelp/logparabola.html>`_ package, however, uses :math:`log_{10}`. If you have
parametrization based on :math:`log_{10}` you can use the
:func:`~gammapy.modeling.models.LogParabolaSpectralModel.from_log10` method.

"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.modeling.models import LogParabolaSpectralModel, Models, SkyModel

energy_bounds = [0.1, 100] * u.TeV
model = LogParabolaSpectralModel(
    alpha=2.3,
    amplitude="1e-12 cm-2 s-1 TeV-1",
    reference=1 * u.TeV,
    beta=0.5,
)
model.plot(energy_bounds)
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(spectral_model=model, name="log-parabola-model")
models = Models([model])

print(models.to_yaml())
